
import gc
import logging
import os
import signal
import socket
import time
from datetime import datetime

from job import Job

log = logging.getLogger(__name__)

class Scheduler:

    DEFAULT_CONFIG = {
      "timeout": 60,
      "heartbeat_key": "mani:heartbeat",
    }

    TRAPPED_SIGNALS = (
        signal.SIGINT,
        signal.SIGTERM,
        signal.SIGQUIT
    )

    def __init__(self, redis, config = {}):
        self.jobs = {}
        self.redis = redis
        self.host = socket.gethostname()
        self.pid = os.getpid()

        self.running = False
        self.stopped = False

        self.config = self.DEFAULT_CONFIG.copy()
        self.config.update(config)

    def add_job(self, period, at, job_func):
        name = job_func.__name__
        if name in self.jobs:
            raise "duplicate job %s" % name

        job = Job(name, period, at, job_func, self.redis, self.config)
        self.jobs[name] = job

    def start(self):
        self.running = True
        self.trap_signals()

        while True:
            if self.stopped: break

            now = self.now()

            jobs = self.jobs_to_run(now)
            for job in jobs:
                job.run(now)

            self.heartbeat(now)

            if self.stopped: break

            self.sleep_until_next_second()

        log.info("stopped")

    def jobs_to_run(self, now):
        return filter(lambda j: j.ready_to_run(now), self.jobs.values())

    def heartbeat(self, now):
        self.redis.hset(self.config["heartbeat_key"], self.heartbeat_field(), now)

    def heartbeat_field(self):
        return "%s##%s" % (self.host, self.pid)

    def now(self):
        return datetime.now()

    def trap_signals(self):
        try:
            for sig in self.TRAPPED_SIGNALS:
                signal.signal(sig, self.stop)
        except ValueError: # for tests to pass (since it runs on a thread)
            log.warning("could not add handlers for trapping signals")

    def stop(self, _signal=None, _frame=None):
        self.stopped = True

    def sleep_until_next_second(self):
        # process gets hot otherwise
        gc.collect()

        now = datetime.now()
        sleeptime = 1.0 - (now.microsecond / 1000000.0)
        time.sleep(sleeptime)





import redis
import time
import urlparse
from datetime import datetime, timedelta

def redis_conn(redis_url):
    info = urlparse.urlparse(redis_url)
    host = info.hostname
    port = info.port
    password = info.password

    return redis.StrictRedis(
        host=host,
        port=port,
        password=password
    )

def to_datetime(ts):
    return datetime.fromtimestamp(float(ts))

def to_timestamp(dt):
    return time.mktime(dt.timetuple())

def next_weekday(wday, dt):
    if dt.weekday() == wday:
        return dt
    while (dt.weekday() != wday):
        dt = dt + timedelta(days = 1)
    return dt




import util

import logging
logging.getLogger(__name__).addHandler(logging.NullHandler())

from scheduler import Scheduler

class Mani:

    def __init__(self, redis_url, config = {}):
        self.redis = util.redis_conn(redis_url)
        self.scheduler = Scheduler(redis=self.redis, config=config)

    def every(self, seconds=0, minutes=0, hours=0, days=0, weeks=0, at=None):

        def inner(job_func):
            period = seconds + minutes * 60
            period += hours * 60 * 60
            period += days * 24 * 60 * 60
            period += weeks * 7 * 24 * 60 * 60
            self.scheduler.add_job(period, at, job_func)

        return inner

    def start(self):
        self.scheduler.start()

    def stop(self):
        self.scheduler.stop()




import unittest
import time
from threading import Thread

from mani import Mani

redis_url = "redis://localhost:6379/"

jobs_ran = []
mani = Mani(redis_url)

@mani.every(seconds=2)
def run_one_job():
    jobs_ran.append("run_one_job")

@mani.every(seconds=5)
def run_two_job():
    jobs_ran.append("run_two_job")

class TestScheduler(unittest.TestCase):

    def test_scheduler(self):
        t = Thread(target=mani.start)
        t.start()

        time.sleep(6)
        mani.stop()
        t.join()

        self.assertTrue(5 <= len(jobs_ran) <= 7)

        one_job_ran = filter(lambda j: j == "run_one_job", jobs_ran)
        self.assertTrue(3 <= len(one_job_ran) <= 4)

        two_job_ran = filter(lambda j: j == "run_two_job", jobs_ran)
        self.assertTrue(1 <= len(two_job_ran) <= 2)


    def tearDown(self):
        mani.redis.delete("mani:job:run_one_job")
        mani.redis.delete("mani:job:run_two_job")




import unittest
from datetime import datetime, timedelta

from mani.job import Job
from mani import util

redis = util.redis_conn("redis://localhost:6379/")

jobs_ran = []
def run_job():
    jobs_ran.append(1)

class TestJob(unittest.TestCase):

    config = {
      "timeout": 60,
      "heartbeat_key": "mani:heartbeat",
    }

    def test_init(self):
        job = Job("test", 1, None, run_job, redis, self.config)
        now = datetime.now()

        self.assertIsNotNone(job)
        self.assertIs(job.is_running(), False)
        self.assertEqual(job.last_ran(now), now - timedelta(seconds=1))

    def test_should_run(self):
        job = Job("test", 1, None, run_job, redis, self.config)
        now = datetime.now()

        self.assertIs(job.ready_to_run(now), True)

    def test_run(self):
        job = Job("test", 1, None, run_job, redis, self.config)
        now = datetime.now()

        self.assertEqual(len(jobs_ran), 0)
        self.assertIs(job.ready_to_run(now), True)
        job.run(now)
        self.assertIs(job.ready_to_run(now), False)
        self.assertEqual(len(jobs_ran), 1)

    def tearDown(self):
        redis.delete("mani:job:test")


if __name__ == '__main__':
    unittest.main()





import re
from datetime import timedelta

import util

class RunAt:
    MINUTE_PATTERN = "(\d\d?):(\d\d?)"
    HOUR_PATTERN = "(\d\d?):(\d\d?):(\d\d?)"
    WEEKDAY_PATTERN = "(...) %s" % HOUR_PATTERN

    PATTERNS = [MINUTE_PATTERN, HOUR_PATTERN, WEEKDAY_PATTERN]

    WEEKDAY_MAP = {
        "mon": 0,
        "tue": 1,
        "wed": 2,
        "thu": 3,
        "fri": 4,
        "sat": 5,
        "sun": 6
    }

    def __init__(self, period, at, now, offset=None):
        self.period = period
        self.at = at
        self.offset = offset
        self.now = now
        self.minute = None
        self.hour = None
        self.wday = None

    def parse(self):

        # weekday pattern
        regex = re.compile(self.WEEKDAY_PATTERN)
        match = re.match(regex, self.at)
        if match:
            self.wday = self.WEEKDAY_MAP.get(match.group(1).lower(), 0)
            self.hour = int(match.group(2))
            self.minute = int(match.group(3))
            return True

        # hour pattern
        regex = re.compile(self.HOUR_PATTERN)
        match = re.match(regex, self.at)
        if match:
            self.hour = int(match.group(1))
            self.minute = int(match.group(2))
            return True

        # minute pattern
        regex = re.compile(self.MINUTE_PATTERN)
        match = re.match(regex, self.at)
        if match:
            self.minute = int(match.group(1))
            return True

    def find_at(self, op_func):
        run_at = op_func(self.offset, timedelta(seconds=self.period))
        if not self.at:
            return run_at

        matched = self.parse()
        if not matched:
            return self.offset

        if self.minute is not None:
            run_at = run_at.replace(minute=self.minute)
        if self.hour is not None:
            run_at = run_at.replace(hour=self.hour)
        if self.wday is not None:
            run_at = util.next_weekday(self.wday, run_at)

        return run_at

    def last_at(self):
        self.offset = self.now - timedelta(seconds=self.period)
        last_ran = self.next_at()
        if last_ran >= self.now:
            last_ran = last_ran - timedelta(seconds=self.period)
        return last_ran

    def next_at(self):
        run_at = self.offset + timedelta(seconds=self.period)
        if not self.at:
            return run_at

        matched = self.parse()
        if not matched:
            return self.offset

        if self.minute is not None:
            run_at = run_at.replace(minute=self.minute)
        if self.hour is not None:
            run_at = run_at.replace(hour=self.hour)
        if self.wday is not None:
            run_at = util.next_weekday(self.wday, run_at)

        return run_at




import logging
import redis_lock

import util
from run_at import RunAt

log = logging.getLogger(__name__)

class Job:
    def __init__(self, name, period, at, func, redis, config):
        self.name = name
        self.period = period
        self.at = at
        self.func = func
        self.redis = redis
        self.running = False
        self.config = config

    def run(self, now):

        lock = redis_lock.Lock(self.redis, self.name, expire=self.config["timeout"])
        if lock.acquire(blocking=False):
            try:
                if not self.ready_to_run(now):
                    return
                log.info("running job %s", self.name)

                self.running = True
                self.set_last_ran(now)

                try:
                    self.func()
                except:
                    log.exception("%s job failed to run!" % self.name)
            finally:
                lock.release()
        else:
            log.info("could not acquire lock for job %s", self.name)

        self.running = False

    def ready_to_run(self, now):
        last_ran = self.last_ran(now)

        run_at = RunAt(self.period, self.at, now, offset=last_ran).next_at()
        log.debug("%s next run is at %s", self.name, run_at)
        if run_at > now or last_ran > now:
            return False

        log.debug("%s run_at: %s, last_ran: %s, now: %s", self.name, run_at, last_ran, now)
        return True

    def last_ran_key(self):
        return "mani:job:%s" % self.name

    def set_last_ran(self, now):
        self.redis.set(self.last_ran_key(), util.to_timestamp(now))

    def last_ran(self, now):
        last_ran = self.redis.get(self.last_ran_key())
        if last_ran:
            return util.to_datetime(last_ran)

        # new job
        last_ran = RunAt(self.period, self.at, now).last_at()
        log.debug("%s new job, last ran would have been at %s", self.name, last_ran)

        return last_ran

    def is_running(self):
        return self.running



#!/usr/bin/env python
from setuptools import setup


setup(name='mani',
      version='0.1',
      packages=['mani'],
      install_requires=['redis',
                        'python-redis-lock'],
      setup_requires=['setuptools'],
      tests_require=['flake8'],
      description='Distributed cron',
      author='Sherin Kurian',
      author_email='sherinkurian.123@gmail.com',
      url='https://github.com/sherinkurian/mani',
      download_url='https://github.com/sherinkurian/mani/tarball/0.1',
      keywords=['cron', 'distributed', 'redis'],
      package_data={'mani.version': ['VERSION']},
      classifiers=['Environment :: Console',
                   'Intended Audience :: Developers',
                   'Operating System :: Unix',
                   'Operating System :: POSIX',
                   'Programming Language :: Python',
                   'Topic :: System',
                   'Topic :: Software Development',
                   'Development Status :: 4 - Beta'])
